#define _BSD_SIGNALS
#include <signal.h>


#include <Inventor/SoDB.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoWWWInline.h>
#include <Inventor/SbBasic.h>
#include <Inventor/Xt/SoXtComponent.h>


#include <Xm/Xm.h>
#include <Xm/SelectioB.h>

#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "URL.h"
#include "WebNode.h"
#include "SoXtWebViewer.h"
#include "sockinet.h"
#include "StringRepos.h"
#include "SceneParser.h"


#define MAXLEN 1024    // buffer length
#define OK_CODE "200"  // status code for OK in HTTP/1.x
#define CRLF  "\15\12"

#undef NDEBUG

/*inline*/ void 
dbMsg( const char* msg )
  {
#ifdef NDEBUG
     cerr << msg << endl;
#endif
  }
    
//!!!!!!!!!!
URL oldURL( "http://www.construct.net:80/worlds/" );
SoXtWebViewer* Viewer;
//!!!!!!!!!

// reqURL  - requested URL
// prevURL - previous URL

SoSeparator* 
download( const URL* reqURL, URL* prevURL )
  {
     *prevURL = reqURL->getFullURL();
     WebNode node = reqURL->getFullURL();
     if( ! node )
       {
	  Viewer->showErrorMsg( "Error", node.getErrMsg() );
	  return NULL;
       }
     ofstream log( "dump.log" );
     log.write( node.getBuffer(), node.getBufSize() );

     SoInput mySceneInput;
     mySceneInput.setBuffer( node.getBuffer(), node.getBufSize() );
     dbMsg( "Scene constructing..." );
     SoSeparator* myGraph = SoDB::readAll( &mySceneInput );
     delete[] node.getBuffer();
     if( myGraph == NULL )
       {
	  Viewer->showErrorMsg( "Construction Error", "Check format" );
	  return NULL;
       }
     dbMsg( "Scene constructed" );
     return myGraph;
  }

SoSeparator*
readFile( const URL* reqURL, URL* )
  {
     SoInput sceneInput;
     
     if( ! sceneInput.openFile( reqURL->getResourceName() ) )
       {
	  Viewer->showErrorMsg( "File Error", "File not found" );
	  return NULL ;
       }
     dbMsg( "Scene constructing..." );
     SoSeparator* myGraph = SoDB::readAll( &sceneInput );
     if( myGraph == NULL )
       {
	  Viewer->showErrorMsg( "Construction Error", "Check format" );
	  return NULL ;
       }
     dbMsg( "Scene constructed" );
     return myGraph;
  }

void
fetchURL( const SbString& url, void* userData, SoWWWInline* node )
  {
     URL* prevURL = ( URL* )userData;	  
     const char* reqURL = url.getString();
#ifdef DEBUG
     cerr << url.getString() << endl; 
#endif
     char buf[MAXLEN];

     if( strncmp( reqURL, "http://", 7 ) )
       {// making up full URL
	  const char* begPtr = prevURL->getFullURL();
	  char* endPtr = strrchr( begPtr, '/' )+1;
	  int len = endPtr - begPtr;
	  memcpy( buf, begPtr, len );
	  strcpy( buf+len, reqURL );
	  reqURL = buf;
       }
     else
       {// saving full URL
	  prevURL->setFullURL( reqURL );
       }
     URL fullURL = reqURL;
     SoSeparator* myGraph = download( &fullURL, prevURL );
     if( !myGraph )
	myGraph = new SoSeparator;
#ifdef DEBUG
     cerr << getpid() <<": myGraph before setChildData(): " << myGraph->getRefCount() << endl;
#endif
     myGraph->ref();
     node->setChildData( myGraph );
     myGraph->unref();
#ifdef DEBUG
     cerr << getpid() << ": myGraph after setChildData(): " << myGraph->getRefCount() << endl;
#endif
  }

const char*
makeFullURL( const char* location, const char* filename )
  {
     static char buf[MAXLEN];

     char* endPtr = strrchr( location, '/' )+1;
     int len = endPtr - location;
     memcpy( buf, location, len );
     strcpy( buf+len, filename );
     return buf;
  }

#define TEX_DIR "/tmp/"

void
addTex( SoGroup* root, StringRep& texDB )
  {
     int chNum = root->getNumChildren();
     for( int idx=0; idx < chNum; idx++ )
       {
	  if( root->getChild( idx )->isOfType( SoTexture2::getClassTypeId() ) )
	    {
	       SoTexture2* tex = ( SoTexture2* )root->getChild( idx );
	       const char* fileName = tex->filename.getValue().getString();
	       if( strcmp( fileName, "" ) )
		 {
		    if( ! strncmp( fileName, TEX_DIR, 5 ) )
		       return;
		    
		    char shortName[128] = TEX_DIR;
		    if( texDB.search( fileName ) )
		      {
			 tex->filename.setValue( strcat( shortName, fileName ) );
			 return;
		      }
		    
		    texDB.add( fileName );
		    
		    if( strncmp( fileName, "http://", 7 ) )
		      {
			 strcat( shortName, fileName );
			 fileName = makeFullURL( oldURL.getFullURL(), fileName );
		      }
		    else
		       strcat( shortName, strrchr( fileName, '/' )+1 );
		    
		    dbMsg( fileName );
		    WebNode node( fileName );
		    if( ! node )
		      {
			 Viewer->showErrorMsg( "Texture Error", node.getErrMsg() );
			 return;
		      }
		    ofstream ofs( shortName );
		    ofs.write( node.getBuffer(), node.getBufSize() );
		    ofs.close();
		    delete[] node.getBuffer();
		    tex->filename.setValue( shortName );
		 }
	       else
		  cerr << "addTex: empty filename\n";
	    }
	  else if( root->getChild( idx )->isOfType( SoGroup::getClassTypeId() ) )
	     addTex( ( SoGroup* )root->getChild( idx ), texDB );
       }
  }

#define FILE_PREF "file:"

void
AddMovingObject( const char* currURL, SoXtFullViewer* viewer )
{
  URL fullURL = currURL;
  SoSeparator *myGraph;
  if( strncmp( currURL, FILE_PREF, strlen( FILE_PREF ) ) )
    myGraph = download( &fullURL, &oldURL );
  else
	myGraph = readFile( &fullURL, &oldURL );
     
     
     if( !myGraph )
	myGraph = new SoSeparator;
#ifdef DEBUG
     cerr << getpid() << ": new myGraph: " << myGraph->getRefCount() << endl;
#endif
     myGraph->ref();
#ifdef DEBUG
     cerr << getpid() << ": myGraph after ref(): " << myGraph->getRefCount() << endl;
#endif
     SoSeparator* old_root = NULL;
     if( viewer->getSceneGraph()->isOfType( SoSeparator::getClassTypeId() ) )
	old_root = ( SoSeparator* )viewer->getSceneGraph();
     else
       {
	  cerr << "viewer->getSceneGraph() doesn't return SoSeparator\n";
	  cerr << "The type is "
	     << viewer->getSceneGraph()->getTypeId().getName().getString() << endl;
	  exit(1);
       }

     // download needed textures
     //     StringRep texDB( myGraph->getNumChildren() );
     //     addTex( myGraph, texDB );

     old_root->addChild( myGraph );
     cerr << getpid() << ": child added\n";
}

void
LoadGraph( const char* currURL, SoXtFullViewer* viewer )
  {
     URL fullURL = currURL;
     SoSeparator *myGraph;
     if( strncmp( currURL, FILE_PREF, strlen( FILE_PREF ) ) )
	myGraph = download( &fullURL, &oldURL );
     else
	myGraph = readFile( &fullURL, &oldURL );
     
     
     if( !myGraph )
	myGraph = new SoSeparator;
#ifdef DEBUG
     cerr << getpid() << ": new myGraph: " << myGraph->getRefCount() << endl;
#endif
     myGraph->ref();
#ifdef DEBUG
     cerr << getpid() << ": myGraph after ref(): " << myGraph->getRefCount() << endl;
#endif
     SoSeparator* old_root = NULL;
     if( viewer->getSceneGraph()->isOfType( SoSeparator::getClassTypeId() ) )
	old_root = ( SoSeparator* )viewer->getSceneGraph();
     else
       {
	  cerr << "viewer->getSceneGraph() doesn't return SoSeparator\n";
	  cerr << "The type is "
	     << viewer->getSceneGraph()->getTypeId().getName().getString() << endl;
	  exit(1);
       }

     // download needed textures
     //     StringRep texDB( myGraph->getNumChildren() );
     //     addTex( myGraph, texDB );

     viewer->setSceneGraph( myGraph );
#ifdef DEBUG
     cerr << getpid() << ": old_root before unref(): " << old_root->getRefCount() << endl;
#endif
     old_root->unref();
#ifdef DEBUG
     cerr << getpid() << ": old_root after unref(): " << old_root->getRefCount() << endl;
#endif
     ParseSceneGraph( myGraph );
#ifdef DEBUG
     cerr << getpid() << ": after ParseSceneGraph\n";
#endif
#ifdef DEBUG
     cerr << getpid() << ": new root at the end: " 
	<< viewer->getSceneGraph()->getRefCount() << endl;
#endif
     dbMsg( "loadGraph ends" );
  }

void
RegisterURLCB()
  {
     SoWWWInline::setFetchURLCallBack( fetchURL, &oldURL );
  }
